// Storage Inspector Lite extension
// storageinspector.js - Handles user events and displays table contents

// Initial Developer: Aaron Pauls
// Contributors:
// See license.txt for source license information.

var StorageInspector = 
{    
    listboxDatabase: null,  // Database list popup menu
    listboxTable: null,     // Table list popup menu
    sCurrentDatabase: null, // Currently selected database
    sCurrentTable: null,    // Currently selected table in the databsae
    aTableData: null,       // Array of table rows (each row is an array of column values)
    aTableHeader: null,     // Array of table headers
    prefService: null,      // Pref service: nsIPrefService
    clipService: null,      // Clipboard service: nsIClipboardHelper
    txtStatus: null,        // Status bar     
    iDebugMode: 0,          // Debug mode

    // Startup: called ONCE during the browser window "load" event
    Startup: function() 
    {
        // Load prefs service
        this.prefService= Components.classes["@mozilla.org/preferences-service;1"]
          .getService(Components.interfaces.nsIPrefService)
          .getBranch("storageinspector.");
        this.GetUserSettings();

        // Load clipboard service
        this.clipService= Components.classes["@mozilla.org/widget/clipboardhelper;1"]
          .getService(Components.interfaces.nsIClipboardHelper);
        
        // Reference page items
        this.listboxDatabase= document.getElementById("dblist");
        this.listboxTable= document.getElementById("tablelist");
        this.txtStatus= this.GetWindowStatusBar();
        TreeTable.Initialize("datatable");
        
        // Observer for storageinspector.* prefs
        //this.prefService.QueryInterface(Components.interfaces.nsIPrefBranch2);
        //this.prefService.addObserver("", this, false);

        this.PopulateDatabaseList();   // Load .sql file list into dropdown 
    },

    // Shutdown: called ONCE during the browser window "unload" event
    Shutdown: function()
    {
        //this.prefService.removeObserver("", this);
        this.prefService= null;
        this.clipService= null;
    },  

    // observe: Subclassed observer function to watch pref changes in StorageInspector
    //observe: function(subject, topic, data)
    //{
    //    if (topic=="nsPref:changed" && (data=="xxx" || data=="yyy")) // If a pref we care about was changed...
    //    { 
    //    }
    //},

    GetUserSettings: function()
    {
        try { this.iDebugMode= this.prefService.getIntPref("debug"); }
        catch (e) { this.iDebugMode= 0; }
    },

    // PopulateDatabaseList: Load list of databases (.sqlite files)
    PopulateDatabaseList: function()
    {
        ClearElement(this.listboxDatabase.firstChild);
        this.listboxDatabase.selectedIndex= -1;
        
        AddDropdownItem("(Unselected)", this.listboxDatabase, true);
        // Get all .sqlite files in the profile directory
        var fileList= Components.classes["@mozilla.org/file/directory_service;1"]
                .getService(Components.interfaces.nsIProperties)
                .get("ProfD", Components.interfaces.nsIFile)
                .directoryEntries; // Get the nsIFile object pointing to the profile directory
        var file;
        var iFileCount=0;
        var iPosition;
        while (fileList.hasMoreElements()) 
        {
            file= fileList.getNext().QueryInterface(Components.interfaces.nsIFile);
            // Should check permissions here
            iPosition= file.leafName.lastIndexOf(".sqlite");
            if (iPosition>0 && iPosition==file.leafName.length-7)
            {
                iFileCount= iFileCount + 1;
                AddDropdownItem(file.leafName, this.listboxDatabase, false);
                if (iFileCount==999) // Limit to under one thousand database files - can increase this if necessary
                   break;
            } 
        }
        StorageInspector.SetStatusMessage("Database list loaded (" + iFileCount + " files)", 1);

        // After database list is loaded, load the table list
        this.PopulateTableList();
    },

    // PopulateTableList: Load list of tables from current database
    PopulateTableList: function()
    {        
        ClearElement(this.listboxTable.firstChild);
        this.listboxTable.selectedIndex= -1;

        if (this.sCurrentDatabase==null) // No database is currently selected
        {
            AddDropdownItem("(N/A)", this.listboxTable, true);
            return;
        }
        try { 
            Database.ConnectToDatabase(this.sCurrentDatabase); 
        } catch (e) 
        { 
            var sTemp= this.sCurrentDatabase;
            this.sCurrentDatabase= null; 
            this.PopulateDatabaseList();
            StorageInspector.SetStatusMessage("Connect to '" + sTemp + "' failed: " + e, 3); 
            return;
        }
       
        try {
            var list= Database.GetTableList();
        } catch (e)
        {
            StorageInspector.SetStatusMessage(e, 3);
        }

        ClearElement(this.listboxTable.firstChild);        
        if (list==null) 
        {
            this.AddDropdownItem("(Empty)", this.listboxTable, true);
            StorageInspector.SetStatusMessage("Loaded '" + this.sCurrentDatabase + "' - (0 tables)", 1);
            return;
        }

        AddDropdownItem("(Unselected)", this.listboxTable, true);

        for (var i=0; i<list.length; i++)
            AddDropdownItem(list[i], this.listboxTable, false);

        StorageInspector.SetStatusMessage("Loaded '" + this.sCurrentDatabase + "' - (" + list.length + " tables)", 1);
    },

    // UserSelectDatabase: Handle user selecting a file from the database dropdown list
    UserSelectDatabase: function()
    {        
        var menuitem= this.listboxDatabase.selectedItem;
        var sLabel;

        if (menuitem!=null)
            sLabel= menuitem.label;
        else
            sLabel= ""; 
        StorageInspector.SetStatusMessage("Selected '" + sLabel + "' database", 1);
        this.SetDatabase(sLabel);
    },

    // UserSelectTable: Handle user selecting a table from the table dropdown list
    UserSelectTable: function()
    {
        var menuitem= this.listboxTable.selectedItem;
        var sTable;

        if (menuitem!=null)
            sTable= menuitem.label;
        else
            sTable= "";

        if (sTable.length==0 || sTable=="(Unselected)" || sTable=="(Empty)" || sTable=="(N/A)") // No item selected by user
        {
            this.sCurrentTable= null;
            TreeTable.ShowTable(false);
            Database.LoadTableData(null);
            this.SelectCommentedItem(this.listboxTable);
            StorageInspector.SetStatusMessage("No table selected", 1);
        }
        else // Valid table selected by user
        {
            this.sCurrentTable= sTable;
            StorageInspector.SetStatusMessage("Table '" + this.sCurrentTable + "' selected", 1);
        }

        try {
            Database.LoadTableData(this.sCurrentTable);
        } catch (e) 
        { 
            StorageInspector.SetStatusMessage(e, 3);
        }
        if (Database.aTableData && Database.aColumns)
            TreeTable.PopulateTableData(Database.aTableData, Database.aColumns);
    },

    // SelectCommentedItem: Select the first item in the list that starts with a '('
    SelectCommentedItem: function(listbox)
    {
        var listboxchild= listbox.firstChild;
        for each (menuitem in listboxchild.childNodes)
            if (menuitem.label.charAt(0)=='(') {
                if (listbox.selectedItem==menuitem)
                    return false; // The item was already selected
                listbox.selectedItem= menuitem;
                listbox.doCommand();
                return true;
            }
        return false;       
    },

    // SetDatabase: Store the database that was selected and fill the table list for that database
    SetDatabase: function(sDatabase) 
    {
        if (sDatabase.length==0 || sDatabase=="(Unselected)") // No item selected by user
        {
            this.SelectCommentedItem(this.listboxDatabase);
            this.sCurrentDatabase= null;            
            StorageInspector.SetStatusMessage("No database selected", 1);
        }
        else // Valid database selected by user
        {
            this.sCurrentDatabase= sDatabase;
            StorageInspector.SetStatusMessage("Database '" + this.sCurrentDatabase + "' selected", 1);
        }
        this.SelectCommentedItem(this.listboxTable);

        this.PopulateTableList();
    },

    // GetWindowStatusBar: Get the statusbar to write notification messages; create one if necessary
    GetWindowStatusBar: function() 
    {
        var statusbar;
        try { 
            statusbar= window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
              .getInterface(Components.interfaces.nsIWebNavigation)
              .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
              .rootTreeItem
              .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
              .getInterface(Components.interfaces.nsIDOMWindow)             
              .document.getElementById("statusbar-display");
            if (!statusbar)
                throw "";
            return statusbar;
        } catch (e) {}
        // If there is no statusbar, add one        
        statusbar= document.createElement("statusbar");
        statusbar.setAttribute("flex","1");
        document.getElementById("statusarea").appendChild(statusbar);
        var statusbarpanel= document.createElement("statusbarpanel");
        statusbarpanel.setAttribute("flex","1");
        statusbar.appendChild(statusbarpanel);
        return statusbarpanel;
    },

    // SetStatusMessage: Display a statusbar message
    SetStatusMessage: function(sMessage, iType)
    {
        // 0= informational       (log only)
        // 1= informational       (status)
        // 2= warning             (status)
        // 3= error               (status + alert)
        // 4= miscellaneous alert (status + alert)

        if (iType>0) {
            if (this.iDebugMode && this.iDebugMode>0)
                dump("StorageInspector: " + sMessage + '\n');
            if (this.txtStatus) 
                this.txtStatus.label= sMessage;
        }
        if (iType>=3)
            alert(sMessage);
    }

};

// ****** table event handling and display ******
var TreeTable = 
{    
    treeTable: null,        // Tree containing listing of current table
    lastRow: null,          // Last row that was clicked on (may be inaccurate if scrolling happened)
    lastCol: null,          // Last col that was clicked on (may be inaccurate if scrolling happened)

    // Initialize: Set up the treeview which will display the table contents
    Initialize: function(sTableID) 
    {        
        this.treeTable= document.getElementById(sTableID);
    },

    // ShowTable: Show/hide any currently displayed table data
    ShowTable: function(bStatus) 
    {
        if (this.treeTable==null) 
            return;
        if (bStatus) // Show the tree
        {
            this.treeTable.childNodes[0].style.display="";
            this.treeTable.childNodes[1].style.display="";
        } else // Hide the tree and remove all of the child rows/columns
        {
            ClearElement(this.treeTable.childNodes[0]);
            ClearElement(this.treeTable.childNodes[1]);
            this.treeTable.childNodes[0].style.display="none";
            this.treeTable.childNodes[1].style.display="none";
        } 
    },
    
    // UserCopyCell: Handle user choosing to copy the cell contents
    UserCopyCell: function()
    {
        var sCell;
        if (this.lastRow && this.lastCol)
        {
            sCell= this.treeTable.view.getCellText(this.lastRow, this.lastCol);
            if (sCell)
                StorageInspector.clipService.copyString(sCell);
        }
    },

    // UserCopyRows: Handle user choosing to copy all currently highlighted rows
    UserCopyRows: function()
    {
        var sel= this.treeTable.view.selection;
        var start= new Object();
        var end= new Object();
        var iRangeCount= sel.getRangeCount();
        var sData="";
        for (var iRange=0; iRange<iRangeCount; iRange++){
          sel.getRangeAt(iRange,start,end);
          for (var iRow=start.value; iRow<=end.value; iRow++){            
            if (sData.length>0)
              sData= sData + "\n";
            sData= sData + this.GetRowData(iRow, "csv");
          }
        }
        if (sData)
            StorageInspector.clipService.copyString(sData);

    },

    // GetRowData: Get the contents of an entire numbered row, using the specified format
    GetRowData: function(iRow, sFormat) 
    {
        var result= null;
        if (sFormat=="csv")
            result="";

        var iColumnCount= 0;
        for (var col in Database.aColumns) 
            iColumnCount= iColumnCount + 1;
        for (var iColumn=0;iColumn<iColumnCount;iColumn++)
            for (var col in Database.aColumns)
                if (Database.aColumns[col][0]==iColumn) {
                    if (sFormat=="csv") {
                        if (iColumn>0)
                            result= result + ", ";
                        if (Database.aColumns[col][2]==3) // String
                            result=result + '"';
                        result= result+ Database.aTableData[iRow][Database.aColumns[col][0]];
                        if (Database.aColumns[col][2]==3) // String
                            result=result + '"';                        
                    }
                }
        return result;       
    },

    // UserTreeClick: Handle the user clicking on the tree
    UserTreeClick: function(ev)
    {
        // This event happens AFTER the window has scrolled (if necessary). This means that
        // if the user clicked on an element that is partially off screen, and the screen scrolls
        // to fully display it, then the mouse may no longer be over the correct element.
        if (ev.button==2) // Right-click
            this.StoreClickLocation(ev);
    },

    // AddTreecol: Add the column to the tree view
    AddTreecol: function(treecols, col, sTooltip, iWidth, bLast) 
    {
        var treecol= document.createElement("treecol");
        treecol.setAttribute("label", col);
        treecol.setAttribute("id", col);
        if (sTooltip!=null)
            treecol.setAttribute("tooltiptext", sTooltip);
        //dump("[" + col + "] width=" + iWidth + "ex\n");
        if (iWidth>0)
            treecol.setAttribute("style", "width:" + iWidth + "ex");
        if (bLast)
            treecol.setAttribute("flex","1");
        treecols.appendChild(treecol); 
    
        if (!bLast) {
            var splitter= document.createElement("splitter");
            splitter.setAttribute("id", "splitter" + col);
            splitter.setAttribute("class", "tree-splitter");        
            splitter.setAttribute("resizebefore", "closest");
            splitter.setAttribute("resizeafter", "farthest");
            treecols.appendChild(splitter); 
        }
    },

    // StoreClickLocation: Store the row/column that the click occurred; used later when copying cell text
    StoreClickLocation: function(ev)
    {
        // http://www.xulplanet.com/references/elemref/ref_tree.html
        if (ev && this.treeTable && ev.type == "click")
        {
            var row = {}, col = {}, obj = {};
            this.treeTable.treeBoxObject.getCellAt(ev.clientX, ev.clientY, row, col, obj);            
            if (row && row.value && col && col.value!=null) // Clicked on a cell
            {
                this.lastRow= row.value;
                this.lastCol= col.value;
            } else
            {
                this.lastRow= null;
                this.lastCol= null;
            }
        }
    },

    // PopulateTableData: Assign our custom treeview
    PopulateTableData: function(aTableData, aColumns)
    {   
        var treecols= this.treeTable.firstChild;
        ClearElement(treecols);
        
        // Order columns (slow method; needs to be reworked)
        var iColumnCount=0;
        var iRow;
        var iWidth, iTotalWidth, iMaxWidth;
        var sTooltip;
        for (var col in aColumns) 
            iColumnCount= iColumnCount + 1;
        for (var iColumn=0;iColumn<iColumnCount;iColumn++)
            for (var col in aColumns)
                if (aColumns[col][0]==iColumn) {
                    iTotalWidth=0;
                    iMaxWidth=0;
                    for (iRow=0;iRow<20 && iRow<aTableData.length;iRow++) {
                        if (aColumns[col][2]==0)
                            iWidth=5;
                        else if (aColumns[col][2]==1 || aColumns[col][2]==2)
                            iWidth=(aTableData[iRow][iColumn] + " ").length;
                        else
                            iWidth=aTableData[iRow][iColumn].length;
                        if (iWidth>0) {
                            iTotalWidth=iTotalWidth + iWidth;
                            if (iWidth>iMaxWidth)
                                iMaxWidth=iWidth;
                        }
                    }
                    if (iRow>0)
                        iTotalWidth=iTotalWidth/iRow;
                    if (iMaxWidth>0)
                        iTotalWidth=(iTotalWidth+iMaxWidth)/2; 
                    iTotalWidth= iTotalWidth+1;
                    // End up with a width half way between the average width and the max width (plus 1)
                    if (iTotalWidth<0)
                        iTotalWidth= 0;
                    else if (iTotalWidth>25)
                        iTotalWidth= 25;
                    sTooltip= col + " (" + GetColumnTypeString(aColumns[col][2]) + ")";
                    this.AddTreecol(treecols, col, sTooltip, iTotalWidth, (iColumn==iColumnCount-1?true:false));
                    break;
                }
        this.treeTable.view= new this.DatabaseTreeView(aTableData,aColumns);
        this.ShowTable(true);
    },

    // DatabaseTreeView: Create a custom nsITreeView - good for displaying large amounts of data
    DatabaseTreeView: function(aTableData, aColumns) 
    {
        // http://kb.mozillazine.org/Sorting_Trees
        this.aTableData          = aTableData; // 2 dimensional array containing table contents
        this.aColumns            = aColumns;   // Column information (index, order, type)

        this.rowCount            = aTableData.length;  // Number of rows in the table

        this.getCellText         = function(row,column) 
        { 
            var sResult;
            try { sResult= this.aTableData[row][this.aColumns[column.id][0]]; }
            catch (e) { return "<" + row + "," + column.id + ">"; }
            return sResult;
        };

        this.setTree             = function(treebox){ this.treebox=treebox; };
        this.isContainer         = function(row){ return false; };
        this.isSeparator         = function(row){ return false; };
        this.isSorted            = function(row){ return false; };
        this.cycleHeader         = function(col){ this.SortColumn(col); }
        this.getLevel            = function(row){ return 0; };
        this.getImageSrc         = function(row,col){ return null; };
        this.getRowProperties    = function(row,props){};
        this.getCellProperties   = function(row,col,props){};
        this.getColumnProperties = function(colid,col,props){};

        this.SortColumn          = function(col)
        {
            var name  = col.id; 
            var index = this.aColumns[name][0];
            var order = this.aColumns[name][1];
            var isnum = ((this.aColumns[name][2]==1 || this.aColumns[name][2]==1)?1:0);

            this.aColumns[name][1] = (order==0)?1:0; // switch order flag

            this.SortTable(this.aTableData,index,order,isnum);  // sort the table
            this.rowCount= this.aTableData.length; // Not the right place for this but...
            this.treeTable.view = new treeView(this.aTableData, this.aColumns, this.rowCount);       
        };

        // This is the actual sorting method, extending the array.sort() method
        this.SortTable           = function(table,col,order,isnum)
        {
            if(isnum){ /* use numeric comparison */
                if(order==0){ /* ascending */
                    this.columnSort= function (a,b){ return (a[col]-b[col]); };
                }
                else{ /* descending */
                    this.columnSort= function (a,b){ return (b[col]-a[col]); };
                }
            }
            else{ /* use string comparison */
                if(order==0){ /* ascending */
                    this.columnSort= function(a,b){
                        return (a[col]<b[col])?-1:(a[col]>b[col])?1:0; };
                }
                else{ /* descending */
                    this.columnSort= function(a,b){
                        return (a[col]>b[col])?-1:(a[col]<b[col])?1:0; };
                }
            }
            // use array.sort(comparer) method
            table.sort(this.columnSort);
        };
    }
};

// GetColumnTypeString: Determine the type of the table column
function GetColumnTypeString(iType)
{
    switch (iType) 
    {
        case 0: return "<null>";
        case 1: return "int";
        case 2: return "float";
        case 3: return "string";
        case 4: return "blob";
        default: return "unknown";
    }
}

// ClearElement: Remove all child elements 
function ClearElement(el) 
{
    while (el.firstChild) 
    { 
      el.removeChild(el.firstChild);
    }
}

// AddDropdownItem: Add a menuitem to the dropdown
function AddDropdownItem(sLabel, dropdown, selectitem) 
{   
    var menuitem = document.createElement("menuitem");
    menuitem.setAttribute("label", sLabel);
    dropdown.firstChild.appendChild(menuitem); // Append the menu item to the "menupopup" element
    if (selectitem==true)
        dropdown.selectedItem= menuitem;
}

window.addEventListener("load", function(e) { StorageInspector.Startup(); }, false);
window.addEventListener("unload", function(e) { StorageInspector.Shutdown(); }, false);
