// Storage Inspector extension
// storagesqlite.js - sqlite database and table interface

// Initial Developer: Aaron Pauls
// Contributors:
// See license.txt for source license information.

var Database =
{
    storageService: null,   // mozIStorageService
    dataConnection: null,   // Connection to current database (mozIStorageConnection)
    aTableData: null,       // Stores 2D array of table data
    aColumns: null,         // Stores array of column information (col#, ordered, type)

    // ConnectToDatabase: Connects and sets dataConnection
    ConnectToDatabase: function(sDatabaseName)
    {   
        var file; // nsIFile

        if (this.storageService==null)
            this.Initialize();

        this.ResetDatabaseConnection();
        file= Components.classes["@mozilla.org/file/directory_service;1"]
                .getService(Components.interfaces.nsIProperties)
                .get("ProfD", Components.interfaces.nsIFile);
        file.append(sDatabaseName);

        this.dataConnection= this.storageService.openDatabase(file);
    },

    // ResetDatabaseConnection: Free the database resource lock (right now this doesn't do much)
    ResetDatabaseConnection: function() 
    {
        if (this.dataConnection!=null)
            this.dataConnection= null; // There appears to be no API to clear a database connection, so I do this       
        this.aTableData= null;
        this.aColumns= null;
    },

    // Get the mozIStorageService object
    Initialize: function()
    {
        // Get database storage service
        this.storageService= Components.classes["@mozilla.org/storage/service;1"]
          .getService(Components.interfaces.mozIStorageService);
    },

    // GetTableList: Retuns an array of all table names in the current database
    GetTableList: function() 
    {
        var stmt;                // mozIStorageStatement (mozIStorageValueArray)
        var aResult= new Array();
        var iTableCount=0;
        try {
            stmt= this.dataConnection.createStatement(
              "SELECT name FROM sqlite_master WHERE type='table' ORDER BY name");
            while (stmt.executeStep()) 
            {
                iTableCount= iTableCount + 1;
                aResult.push(stmt.getString(0));
                if (iTableCount==9999) //// Limit to under ten thousand tables - can increase this if necessary
                    break;
            }
        } catch (e) {
            throw "Unable to get master list: " + e;
        }        
        aResult.push("sqlite_master");
        return aResult;
    },

    // LoadTableData: Retrieve and save the table data from the storage object
    LoadTableData: function(sTable)
    {
        var stmt;                // mozIStorageStatement (mozIStorageValueArray)
        var iColumn;
        var iRow;
        var iType;
        var aTemp;
        var cell;
        var bResult; 

        if (sTable==null)
        {
            this.ResetDatabaseConnection();
            return;
        }           

        this.aTableData= null;
        this.aColumns= null;        
        bResult= false;

        try {
            stmt= this.dataConnection.createStatement(
              "SELECT * FROM " + sTable);

            bResult= stmt.executeStep(); // We must run this once before we can get the column info
            this.aColumns=new Array();

            for (iColumn=0;iColumn<stmt.columnCount;iColumn++) {
                iType= stmt.getTypeOfIndex(iColumn);
                aTemp= new Array();
                aTemp.push(iColumn);   // Index into the table array
                aTemp.push(0);         // Order
                aTemp.push(iType);     // Type; 0=null, 1=int, 2=float, 3=string, 4=blob
                //aTemp.name= stmt.getColumnName(iColumn);
                this.aColumns[stmt.getColumnName(iColumn)]= aTemp;  // Add this column information using the named column
                if (iColumn==999) // Limit to under one thousand columns - can increase this if necessary
                    break; 
            }
        } catch (e) 
        { 
            throw "Table '" + this.sCurrentTable + "' column load failed; " + e;
        }

        iRow= 0;
        try {
            this.aTableData= new Array();
            while (bResult) {
                aTemp= new Array();
                for (iColumn=0;iColumn<stmt.columnCount;iColumn++) {
                    iType= stmt.getTypeOfIndex(iColumn);
                    switch (iType) {
                        case 0: cell= "<null>";                      break;
                        case 1: cell= stmt.getInt32(iColumn);        break;
                        case 2: cell= stmt.getDouble(iColumn);       break;
                        case 3: cell= stmt.getString(iColumn);       break;
                        case 4: cell= stmt.getString(iColumn);       break; // BLOB
                        default: sData= "<unknown>"; 
                    }
                    aTemp.push(cell);
                }
                iRow= iRow + 1;
                this.aTableData.push(aTemp);
                if (iRow==999999) // Limit to under one million rows - can increase this if necessary
                    break;
                bResult= stmt.executeStep();
            }
        } catch (e) 
        { 
            throw "Table '" + this.sCurrentTable + "' data load failed; " + e;
        }
    }

};